unit CalendarDrawer;

interface
uses
  Windows, SysUtils, DateUtils, Controls, Math, Types, Graphics,
  //
  stdGraphics, stdInt;
type
  TYear               = 1..9999;
  TMonth              = 1..12;
  TDay                = 1..31;
  TWeek               = (wSun, wMon, wTue, wWed, wThu, wFri, wSat);
  TWeekCount          = 0..6;
  TDateElement        = (deSunday, deSaturday, deHoliday, deHighlightDay, deToday);
  TDateElements       = set of TDateElement;
  TDays               = set of TDay;

  TCalendarCellSize   = record
    YearSize          : TSize;
    MonthSize         : TSize;
    WeekSize          : array[TWeekCount] of TSize;
  end;

  TCalendarCellRect   = record
    YearRect          : TRect;
    MonthRect         : TRect;
    WeekRect          : array[TWeekCount] of TRect;
  end;

  TCalendarDrawerBase = class
  private
    mYear             : TYear;
    mMonth            : TMonth;
    mHolydays         : TDays;
    mHighlightDays    : TDays;
    mFirstDayOfWeek   : TWeek;
    mFont             : TFont;
  protected
    function  MeasureYearCell(Canvas: TCanvas; AYear: TYear): TSize; virtual;
    function  MeasureMonthCell(Canvas: TCanvas; AMonth: TMonth): TSize; virtual;
    function  MeasureWeeksCell(Canvas: TCanvas; AWeekOf: TWeek): TSize; virtual;
    function  MeasureSizeToRect(Canvas: TCanvas;CellSize: TCalendarCellSize): TCalendarCellRect; virtual;
    function  MeasureRect(Canvas: TCanvas;var CellRect: TCalendarCellRect): TRect; virtual;

    procedure DoDrawYear(Canvas: TCanvas;Rect: TRect; AYear: TYear); virtual;
    procedure DoDrawMonth(Canvas: TCanvas;Rect: TRect; AMonth: TMonth); virtual;
    procedure DoDrawWeek(Canvas: TCanvas;Rect: TRect; AWeek: TWeek); virtual;
    procedure DoDrawDay(Canvas: TCanvas;Rect: TRect; ADate: TDate; DateElement: TDateElements); virtual;
    procedure DoDrawTrailingDay(Canvas: TCanvas;Rect: TRect; ADate: TDate; DateElement: TDateElements); virtual;
    procedure DoDrawText(Canvas: TCanvas; Rect: TRect; AString: String); virtual;
    procedure DoDrawCalender(Canvas: TCanvas;CellRect: TCalendarCellRect); virtual;

    function  GetHolidays(): TDays; virtual;
    function  GetHighlightDays(): TDays; virtual;

    function  GetCalculateSyunbun(): TDay;
    function  GetDayElements(ADate: TDate): TDateElements; virtual;
  public
    constructor Create(); overload;
    constructor Create(AYear: TYear; AMonth: TMonth); overload; virtual;
    destructor Destroy; override;
    procedure SetMonth(Year: TYear; Month: TMonth);

    property  Year            : TYear     read  mYear;
    property  Month           : TMonth    read  mMonth;
    property  Holydays        : TDays     read  mHolydays         write mHolydays;
    property  HighlightDays   : TDays     read  mHighlightDays    write mHighlightDays;
    property  FirstDayOfWeek  : TWeek     read  mFirstDayOfWeek   write mFirstDayOfWeek default wSun;
    property  Font            : TFont     read  mFont;
  end;

  TCustomCalenderDrawer = class(TCalendarDrawerBase)
  public
    procedure DrawBitmap(ABitmap: TBitmap);
    procedure Draw(ACanvas: TCanvas; X, Y, Width, Height: Integer); overload;
    procedure Draw(ACanvas: TCanvas; X, Y: Integer); overload;
    procedure Draw(ACanvas: TCanvas); overload;
    function  GetSize(ACanvas: TCanvas): TSize;
  end;

  TCalendarColor = class
  private
    mColor        : TColor;
    mFontStyle    : TFontStyles;
    mShadowColor  : TColor;
    mShadowPoint  : TPoint;
  public
    constructor Create(Color, ShadowColor: TColor; FontStyle: TFontStyles = []); overload;
    procedure Assign(Source: TCalendarColor);
    property  Color       : TColor      read  mColor        write mColor;
    property  FontStyle   : TFontStyles read  mFontStyle    write mFontStyle;
    property  ShadowColor : TColor      read  mShadowColor  write mShadowColor;
    property  ShadowPoint : TPoint      read  mShadowPoint  write mShadowPoint;
  end;

  TCalendarDrawer = class(TCustomCalenderDrawer)
  private
    mYearColor        : TCalendarColor;
    mMonthColor       : TCalendarColor;
    mWeekdayColor     : TCalendarColor;
    mSundayColor      : TCalendarColor;
    mSaturdayColor    : TCalendarColor;
    mHolydayColor     : TCalendarColor;
    mHighlightDayColor: TCalendarColor;
    mTodayColor       : TCalendarColor;
    mTrailingDayLight : Byte;
    mDrawColor        : TCalendarColor;
  protected
    property  DrawColor : TCalendarColor  read  mDrawColor write mDrawColor;
    procedure SetColorData(ColorData: TCalendarColor);
    procedure ResetProc();
    function MeasureSizeToRect(Canvas: TCanvas;
      CellSize: TCalendarCellSize): TCalendarCellRect; override;
    procedure DoPreDrawDay(Canvas: TCanvas; DateElement: TDateElements); virtual;
    procedure DoDrawDay(Canvas: TCanvas; Rect: TRect; ADate: TDate;
      DateElement: TDateElements); override;
    procedure DoDrawMonth(Canvas: TCanvas; Rect: TRect; AMonth: TMonth);
      override;
    procedure DoDrawText(Canvas: TCanvas; Rect: TRect; AString: String);
      override;
    procedure DoDrawTrailingDay(Canvas: TCanvas; Rect: TRect; ADate: TDate;
      DateElement: TDateElements); override;
    procedure DoDrawWeek(Canvas: TCanvas; Rect: TRect; AWeek: TWeek);
      override;
    procedure DoDrawYear(Canvas: TCanvas; Rect: TRect; AYear: TYear);
      override;
    procedure DoDrawCalender(Canvas: TCanvas; CellRect: TCalendarCellRect);
      override;

  public
    constructor Create(AYear: TYear; AMonth: TMonth); overload; override;
    destructor Destroy; override;

    property  YearColor        : TCalendarColor read  mYearColor          write mYearColor;
    property  MonthColor       : TCalendarColor read  mMonthColor         write mMonthColor;
    property  WeekdayColor     : TCalendarColor read  mWeekdayColor       write mWeekdayColor;
    property  SundayColor      : TCalendarColor read  mSundayColor        write mSundayColor;
    property  SaturdayColor    : TCalendarColor read  mSaturdayColor      write mSaturdayColor;
    property  HolydayColor     : TCalendarColor read  mHolydayColor       write mHolydayColor;
    property  HighlightDayColor: TCalendarColor read  mHighlightDayColor  write mHighlightDayColor;
    property  TodayColor       : TCalendarColor read  mTodayColor         write mTodayColor;
    property  TrailingDayLight : Byte           read  mTrailingDayLight   write mTrailingDayLight default 40;
  end;

const
  WeekString  : array[TWeekCount] of String =
    ('','','','','','','y');

function NextDayOfWeek(Week: TWeek): TWeek;
function PrevDayOfWeek(Week: TWeek): TWeek;
function GetWeekCount(BaseDayOfWeek: TWeek;Year: TYear; Month: TMonth): Integer;
function GetDayInWeek(Year: TYear; Month: TMonth;Index: Integer; AWeekOf: TWeek): TDay;
function GetDayInWeeks(Year: TYear; Month: TMonth;AWeekOf: TWeek): TDays;

implementation

uses Classes;

{ TCalendarDrawerBase }

constructor TCalendarDrawerBase.Create();
var y, m, d: Word;
begin
  DecodeDate(Now, y, m, d);
  Create(y, m);
end;

constructor TCalendarDrawerBase.Create(AYear: TYear; AMonth: TMonth);
begin
  inherited Create;

  mFont           := TFont.Create;
  // ftHgl̐ݒ
  mFirstDayOfWeek := wSun;
  mYear           := AYear;
  mMonth          := AMonth;
  mHolydays       := GetHolidays;
  mHighlightDays  := GetHighlightDays;
end;

destructor TCalendarDrawerBase.Destroy;
begin
  mFont.Free;
  inherited;
end;

//// \bh

procedure TCalendarDrawerBase.SetMonth(Year: TYear; Month: TMonth);
begin
  mYear     := Year;
  mMonth    := Month;
  mHolydays := GetHolidays;
end;

//// veNg\bh

// W[֐

function TCalendarDrawerBase.MeasureYearCell(Canvas: TCanvas;
  AYear: TYear): TSize;
begin
  Result := Canvas.TextExtent(Format('%4dN', [AYear]));
end;

function TCalendarDrawerBase.MeasureMonthCell(Canvas: TCanvas;
  AMonth: TMonth): TSize;
begin
  Result := Canvas.TextExtent(Format('%2d', [AMonth]));
end;

function TCalendarDrawerBase.MeasureWeeksCell(Canvas: TCanvas;
  AWeekOf: TWeek): TSize;
begin
  Result := Canvas.TextExtent(WeekString[Integer(AWeekOf)]);
  Inc(Result.cx, 2);
  Inc(Result.cy, 2);
end;

function TCalendarDrawerBase.MeasureSizeToRect(Canvas: TCanvas;
  CellSize: TCalendarCellSize): TCalendarCellRect;
var
  i, l: Integer;
begin
  // N`̈ݒ
  Result.YearRect     :=
        Bounds(0, 0, CellSize.YearSize.cx,
        Max( CellSize.YearSize.cy, CellSize.MonthSize.cy));
  // `̈ݒ
  Result.MonthRect    :=
        Bounds(Result.YearRect.Right + 1, 0, CellSize.MonthSize.cx,
        Result.YearRect.Bottom);
  // T`̈ݒ
  l := 0;
  for i := Low(TWeekCount) to High(TWeekCount) do begin
    Result.WeekRect[i]:=
        Bounds(l, Result.MonthRect.Bottom + 1,
        CellSize.WeekSize[i].cx, CellSize.WeekSize[i].cy);
    l := Result.WeekRect[i].Right;
  end;
end;

function TCalendarDrawerBase.MeasureRect(
  Canvas: TCanvas;var CellRect: TCalendarCellRect): TRect;
var
  Size: TCalendarCellSize;
  i   : Integer;
  w   : TWeek;
  c   : Integer;
begin
  Canvas.Font       := Font;
  Size.YearSize     := MeasureYearCell(Canvas, Year);
  Size.MonthSize    := MeasureMonthCell(Canvas, Month);
  w                 := FirstDayOfWeek;
  for i := Low(TWeekCount) to High(TWeekCount) do begin
    Size.WeekSize[i]:= MeasureWeeksCell(Canvas, w);
    w               := NextDayOfWeek(w);
  end;
  CellRect          := MeasureSizeToRect(Canvas, Size);

  // g[^TCY̎擾
  Result            := CellRect.YearRect;
  UnionRect(Result, Result, CellRect.MonthRect);
  c := GetWeekCount(FirstDayOfWeek, Year, Month) + 1;
  for i := Low(TWeekCount) to High(TWeekCount) do
    with CellRect.WeekRect[i] do
      UnionRect(Result, Result,
        Rect(Left, Top, Right, Top + (Bottom - Top) * c));
end;

// `

procedure TCalendarDrawerBase.DoDrawYear(Canvas: TCanvas;
  Rect: TRect; AYear: TYear);
begin
  DoDrawText(Canvas, Rect, Format('%4dN', [AYear]));
end;

procedure TCalendarDrawerBase.DoDrawMonth(Canvas: TCanvas;
  Rect: TRect; AMonth: TMonth);
begin
  DoDrawText(Canvas, Rect, Format('%2d', [AMonth]));
end;

procedure TCalendarDrawerBase.DoDrawWeek(Canvas: TCanvas;
  Rect: TRect; AWeek: TWeek);
begin
  DoDrawText(Canvas, Rect, WeekString[Integer(AWeek)]);
end;

procedure TCalendarDrawerBase.DoDrawDay(Canvas: TCanvas;
  Rect: TRect; ADate: TDate; DateElement: TDateElements);
begin
  DoDrawText(Canvas, Rect, IntToStr(DayOf(ADate)));
end;

procedure TCalendarDrawerBase.DoDrawTrailingDay(Canvas: TCanvas;
  Rect: TRect; ADate: TDate; DateElement: TDateElements);
begin
  DoDrawText(Canvas, Rect, IntToStr(DayOf(ADate)));
end;

procedure TCalendarDrawerBase.DoDrawCalender(Canvas: TCanvas;
  CellRect: TCalendarCellRect);
var
  i : Integer;
  d : TDate;
  w : TWeek;
begin
  //// N
  DoDrawYear(Canvas, CellRect.YearRect, Year);
  //// 
  DoDrawMonth(Canvas, CellRect.MonthRect, Month);
  //// T
  w := FirstDayOfWeek;
  for i := Low(TWeekCount) to High(TWeekCount) do begin
    DoDrawWeek(Canvas, CellRect.WeekRect[i], w);
    w   := NextDayOfWeek(w);
  end;
  //// 
  d := EncodeDate(Year, Month, 1);
  // O̓j܂Ŗ߂
  while TWeek(DayOfWeek(d) - 1) <> FirstDayOfWeek do
    d := d - 1;
//  d := d + 1;
  // t̕`揈
  repeat
    for i := Low(TWeekCount) to High(TWeekCount) do begin
      OffsetRect(CellRect.WeekRect[i], 0,
          CellRect.WeekRect[i].Bottom - CellRect.WeekRect[i].Top);
      if MonthOf(d) = Month then
        DoDrawDay(Canvas, CellRect.WeekRect[i], d,
          GetDayElements(d))
      else
        DoDrawTrailingDay(Canvas, CellRect.WeekRect[i], d,
          GetDayElements(d));
      d := d + 1;
    end;
  until MonthOf(d) <> Month;
end;

procedure TCalendarDrawerBase.DoDrawText(Canvas: TCanvas;
  Rect: TRect; AString: String);
begin
  Canvas.TextRect(Rect, Rect.Left, Rect.Top, AString);
end;

// ̑


function TCalendarDrawerBase.GetHolidays: TDays;
// ݂̃y[W
// http://koyomi.vis.ne.jp/
const
  JpnHolidays : array[1..12] of TDays = (
      [1],      // 1
      [11],     // 2
      [],       // 3
      [29],     // 4
      [3, 5],   // 5
      [],       // 6
      [],       // 7
      [],       // 8
      [],       // 9
      [],       // 10
      [3, 23],  // 11
      [23]      // 12
    );
var i : Integer;
begin
  Result := JpnHolidays[Month];
  //// jǉ
  case Month of
    1: // l̓ǉ
      if Year <= 1999 then // 1999NȑO15Œ
        Include(Result, 15)
      else
        Include(Result, GetDayInWeek(Year, Month, 2, wMon));
    3: // t̓ǉ
      Include(Result, GetCalculateSyunbun());
    7: // C̓ǉ
      if Year <= 2002 then // 2002NȑO20Œ
        Include(Result, 20)
      else
        Include(Result, GetDayInWeek(Year, Month, 3, wMon));
    9: begin
      // H̓ǉ
      Include(Result, GetCalculateSyunbun());
      // hV̓ǉ
      if Year <= 2002 then // 2002NȑO15Œ
        Include(Result, 15)
      else
        Include(Result, GetDayInWeek(Year, Month, 3, wMon));
    end;
    10: // ̈̓ǉ
      if Year <= 1999 then // 1999NȑO10Œ
        Include(Result, 10)
      else
        Include(Result, GetDayInWeek(Year, Month, 1, wMon));
  end;
  // UւxThCb`x̒ǉ
  for i := Low(TDay) to High(TDay) do
    if i in Result then begin
      // Uւx
      if DayOfWeek(EncodeDate(Year, Month, i)) = 1 then Include(Result, i + 1);
      // ThCb`x̒ǉ
      if (i - 2) in Result then Include(Result, i - 1);
    end;
end;

// nCCg\ׂtԂ
function TCalendarDrawerBase.GetHighlightDays: TDays;
begin
  Result := [];
end;

// t܂͏H̓tԂ
function TCalendarDrawerBase.GetCalculateSyunbun: TDay;
const
  par1 : array[0..3] of Single = (19.8277, 20.8357, 20.8431, 21.8510);
  par2 : array[0..3] of Single = (22.2588, 23.2588, 23.2488, 24.2488);
  basepar = 0.242194;
var
  p1, tyear: Integer;
begin
  if Month in [3, 9] then begin;
    case Year of
      1851..1899: p1 := 0;
      1900..1979: p1 := 1;
      1980..2099: p1 := 2;
      2100..2150: p1 := 3;
    else
      p1 := -1;
    end;
    tyear := Year - IfThen(p1 < 2, 1983, 1980);

    if p1 > -1 then
      Result := Floor(
        IfThen(Month = 3, par1[p1], par2[p1]) + basepar *
        tyear - Floor((tyear + 0.1) / 4))
    else
      Result := IfThen(Month = 3, 21, 23); // AȂ̂ɂĂ
  end else
    raise Exception.Create('̌ɏHяt͂܂B');
end;

function TCalendarDrawerBase.GetDayElements(ADate: TDate): TDateElements;
begin
  // yǂ
  case DayOfWeek(ADate) of
    1: Result := [deSunday];
    7: Result := [deSaturday];
  else
    Result    := [];
  end;
  if MonthOf(ADate) = Month then begin
    // jǂ
    if DayOf(ADate) in Holydays then
      Include(Result, deHoliday);
    // nCCgǂ
    if DayOf(ADate) in HighlightDays then
      Include(Result, deHighlightDay);
    // ǂ
    if SameDate(ADate, Date) then
      Include(Result, deToday);
  end;
end;

{ TCustomCalenderDrawer }

procedure TCustomCalenderDrawer.DrawBitmap(ABitmap: TBitmap);
var
  rc    : TRect;
  cell  : TCalendarCellRect;
begin
  rc := MeasureRect(ABitmap.Canvas, cell);
  ABitmap.Width := rc.Right;
  ABitmap.Height:= rc.Bottom;
  DoDrawCalender(ABitmap.Canvas, cell);
end;

procedure TCustomCalenderDrawer.Draw(ACanvas: TCanvas;
  X, Y, Width, Height: Integer);
var
  bitmap: TBitmap;
begin
  bitmap:= TBitmap.Create;
  try
    DrawBitmap(bitmap);
    ACanvas.CopyRect(Bounds(X, Y, Width, Height), ACanvas, Rect(0, 0, Width, Height));
  finally
    bitmap.Free;
  end;
end;

procedure TCustomCalenderDrawer.Draw(ACanvas: TCanvas);
var
  cell: TCalendarCellRect;
begin
  MeasureRect(ACanvas, cell);
  DoDrawCalender(ACanvas, cell);
end;

procedure TCustomCalenderDrawer.Draw(ACanvas: TCanvas; X, Y: Integer);
var
  bitmap: TBitmap;
begin
  bitmap:= TBitmap.Create;
  try
    bitmap.Transparent := true;
    DrawBitmap(bitmap);
    ACanvas.Draw(X, Y, bitmap);
  finally
    bitmap.Free;
  end;
end;

function TCustomCalenderDrawer.GetSize(ACanvas: TCanvas): TSize;
var
  cell: TCalendarCellRect;
  po  : TPoint;
begin
  po    := MeasureRect(ACanvas, cell).BottomRight;
  Result.cx := po.X;
  Result.cy := po.Y;
end;

{ TCalendarColor }

constructor TCalendarColor.Create(Color, ShadowColor: TColor; FontStyle: TFontStyles);
begin
  Create;

  mColor      := Color;
  mShadowColor:= ShadowColor;
  mShadowPoint:= Point(2, 2);
  mFontStyle  := FontStyle;
end;

procedure TCalendarColor.Assign(Source: TCalendarColor);
begin
  if Source.Color <> clNone then
    Color       := Source.Color;
  if Source.ShadowColor <> clNone then
    ShadowColor := Source.ShadowColor;
  ShadowPoint   := Source.ShadowPoint;
  FontStyle     := Source.FontStyle;
end;

{ TCalendarDrawer }

constructor TCalendarDrawer.Create(AYear: TYear; AMonth: TMonth);
begin
  inherited;

  mYearColor        := TCalendarColor.Create(clLime, clGray);
  mMonthColor       := TCalendarColor.Create(clCream, clOlive);
  mWeekdayColor     := TCalendarColor.Create(clBlack, clWhite);
  mSundayColor      := TCalendarColor.Create(clRed, clMaroon);
  mSaturdayColor    := TCalendarColor.Create(clBlue, clNavy);
  mHolydayColor     := TCalendarColor.Create(clFuchsia, clPurple, [fsItalic]);
  mHighlightDayColor:= TCalendarColor.Create(clWhite, clBlack, [fsBold]);
  mTodayColor       := TCalendarColor.Create(clLime, clSilver, [fsUnderline]);
  mTrailingDayLight := $A0;
  mDrawColor        := TCalendarColor.Create();

end;

destructor TCalendarDrawer.Destroy;
begin
  YearColor.Free;
  MonthColor.Free;
  WeekdayColor.Free;
  SundayColor.Free;
  SaturdayColor.Free;
  HolydayColor.Free;
  HighlightDayColor.Free;
  TodayColor.Free;
  DrawColor.Free;

  inherited;
end;

// veNg\bh

procedure TCalendarDrawer.DoDrawCalender(Canvas: TCanvas;
  CellRect: TCalendarCellRect);
begin
  ResetProc;
  inherited;
end;

procedure TCalendarDrawer.SetColorData(ColorData: TCalendarColor);
begin
  if ColorData.Color <> clNone then
    DrawColor.Color       := ColorData.Color;
  if ColorData.ShadowColor <> clNone then
    DrawColor.ShadowColor := ColorData.ShadowColor;
  DrawColor.FontStyle     := DrawColor.FontStyle + ColorData.FontStyle;
  if ColorData.ShadowPoint.X +
    ColorData.ShadowPoint.Y <> 0 then
    DrawColor.ShadowPoint := ColorData.ShadowPoint;
end;

procedure TCalendarDrawer.DoPreDrawDay(Canvas: TCanvas;
  DateElement: TDateElements);
  function IsOpt(Element: TDateElement): Boolean;
  begin
    Result := Element in DateElement;
  end;
begin
  // ̂̂̂قDxƂɂȂ
  SetColorData(WeekdayColor);
  if IsOpt(deSunday) then
    SetColorData(SundayColor)
  else if IsOpt(deSaturday) then
    SetColorData(SaturdayColor);

  if IsOpt(deHighlightDay) then
    SetColorData(HighlightDayColor);
  if IsOpt(deHoliday) then
    SetColorData(HolydayColor);
  if IsOpt(deToday) then
    SetColorData(TodayColor);
end;

procedure TCalendarDrawer.DoDrawDay(Canvas: TCanvas; Rect: TRect;
  ADate: TDate; DateElement: TDateElements);
begin
  DoPreDrawDay(Canvas, DateElement);
  inherited;
end;

procedure TCalendarDrawer.DoDrawTrailingDay(Canvas: TCanvas; Rect: TRect;
  ADate: TDate; DateElement: TDateElements);
  function ShiftColor(AColor: TColor): TColor;
  var
    c: array[0..2] of Byte;
    i: Integer;
  begin
    DivideRGB(AColor, c[0] ,c[1] ,c[2]);
    for i := 0 to High(c) do
      c[i] :=
        EnsureRange(c[i] + TrailingDayLight, 0, 255);
    Result := RGB(c[0], c[1], c[2]);
  end;

begin
  DoPreDrawDay(Canvas, DateElement);
  DrawColor.Color       := ShiftColor(DrawColor.Color);
  DrawColor.ShadowColor := ShiftColor(DrawColor.ShadowColor);
  inherited;
end;

procedure TCalendarDrawer.DoDrawYear(Canvas: TCanvas; Rect: TRect;
  AYear: TYear);
begin
  SetColorData(YearColor);
  inherited;
end;

procedure TCalendarDrawer.DoDrawMonth(Canvas: TCanvas; Rect: TRect;
  AMonth: TMonth);
begin
  SetColorData(MonthColor);
  inherited;
end;

procedure TCalendarDrawer.DoDrawWeek(Canvas: TCanvas; Rect: TRect;
  AWeek: TWeek);
begin
  SetColorData(WeekdayColor);
  case AWeek of
    wSun: SetColorData(SundayColor);
    wSat: SetColorData(SaturdayColor);
  end;
  inherited;
end;

procedure TCalendarDrawer.DoDrawText(Canvas: TCanvas; Rect: TRect;
  AString: String);
const
  Flags : Integer = DT_CENTER or DT_VCENTER or DT_SINGLELINE or DT_NOCLIP
    or DT_NOPREFIX;
begin
  // 
  Canvas.Font.Assign(Font);
  Canvas.Font.Style := DrawColor.FontStyle;
  // `
  if (DrawColor.ShadowPoint.X + DrawColor.ShadowPoint.Y = 0) or (DrawColor.ShadowColor = clNone) then begin
    Canvas.Font.Color := DrawColor.Color;
    SetBkMode(Canvas.Handle, TRANSPARENT);
    DrawText(Canvas.Handle, PChar(AString), -1, Rect, Flags);
  end else
    DrawTextShadow(Rect, Canvas, PChar(AString), Flags,
         DrawColor.ShadowPoint.X, DrawColor.ShadowPoint.Y,
         DrawColor.Color, DrawColor.ShadowColor);
  ResetProc;
end;

function TCalendarDrawer.MeasureSizeToRect(Canvas: TCanvas;
  CellSize: TCalendarCellSize): TCalendarCellRect;
var i : Integer;
begin
  Inc(CellSize.YearSize.cx, YearColor.ShadowPoint.X);
  Inc(CellSize.YearSize.cy, YearColor.ShadowPoint.Y);
  Inc(CellSize.MonthSize.cx, MonthColor.ShadowPoint.X);
  Inc(CellSize.MonthSize.cy, MonthColor.ShadowPoint.Y);
  for i := Low(TWeekCount) to High(TWeekCount) do begin
    Inc(CellSize.WeekSize[i].cx, WeekdayColor.ShadowPoint.X);
    Inc(CellSize.WeekSize[i].cy, WeekdayColor.ShadowPoint.Y);
  end;
  Result := inherited MeasureSizeToRect(Canvas, CellSize);
end;

procedure TCalendarDrawer.ResetProc;
begin
  DrawColor.Color       := clNone;
  DrawColor.ShadowColor := clNone;
  DrawColor.ShadowPoint := Point(0, 0);
  DrawColor.FontStyle   := [];
end;

////// ̑֐
function NextDayOfWeek(Week: TWeek): TWeek;
begin
  if Week = wSat then
    Result := wSun
  else
    Result := Succ(Week);
end;

function PrevDayOfWeek(Week: TWeek): TWeek;
begin
  if Week = wSun then
    Result := wSat
  else
    Result := Pred(Week);
end;

/// w茎TԂ邩ׂ
function GetWeekCount(BaseDayOfWeek: TWeek;Year: TYear; Month: TMonth): Integer;
var
  sd, fd  : TDate;
  count   : Integer;
begin
  // ŏ̓Ō̓𓾂
  sd      := StartOfAMonth(Year, Month);
  fd      := EndOfAMonth(Year, Month);
  // O̎wj܂Ŗ߂
  while TWeek(DayOfWeek(sd) - 1) <> BaseDayOfWeek do
    sd    := sd - 1;
  // T𓾂
  count   := DaysBetween(sd, fd) + 1; // ȂȂ̂Ł{1
  case count of
    0 .. 7 * 4:
      Result := 4;
    7 * 4 + 1 .. 7 * 5:
      Result := 5;
    else
      Result := 6;
  end;
end;

// nTڂ̎wj̓t𓾂
function GetDayInWeek(Year: TYear; Month: TMonth;Index: Integer; AWeekOf: TWeek): TDay;
begin
  Result := 7 - DayOfWeek(StartOfAMonth(Year, Month)) + Integer(AWeekOf) + 1 +
    7 * (Index - 1) + 1;
end;

// ̎wj̏ŴԂ
function GetDayInWeeks(Year: TYear; Month: TMonth;AWeekOf: TWeek): TDays;
var
  i : Integer;
begin
  for i := 1 to NthDayOfWeek(EndOfAMonth(Year, Month)) do
    Include(Result, GetDayInWeek(Year, Month, i, AWeekOf));
end;

end.

